from importlib.metadata import version
import os

# Specify the version of the IonQ API that this SDK targets.
API_VERSION = "0.4"

# Get the version of the current package
SDK_VERSION = version(os.path.basename(os.path.dirname(__file__)))

# Construct the default API URL by incorporating the specified API version into the base URL.
DEFAULT_API_URL = f"https://api.ionq.co/v{API_VERSION}"



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Optional,
    List,
    Union,
)

import networkx as nx

import matplotlib.pyplot as plt

from ionq.schemas.job import (
    JobRequest,
    WorkloadInput,
    JobType,
    OptimizationWorkload,
    OptimizationMethod,
    OptimizationResult as OptimizationResultSchema,
    JobProgress,
)

from .base import Result, Job, Workload

from .quantum_function import QuantumFunction


if TYPE_CHECKING:
    from .backend import Backend


class OptimizationResult(Result):
    def __init__(
        self,
        minimum_value: float,
        optimal_parameters: List[float],
        progress: JobProgress,
        optimal_bitstrings: Optional[List[str]] = None,
    ):
        self.minimum_value = minimum_value
        self.optimal_parameters = optimal_parameters
        self.progress = progress.progress
        self.optimal_bitstrings = optimal_bitstrings

    def plot_results(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 6))

        x = [item.iteration for item in self.progress if item.iteration is not None]
        y = [item.value for item in self.progress if item.value is not None]
        ax.plot(x, y, label="Optimization Progress")

        # Mark the minimum energy point
        min_energy = self.minimum_value
        min_energy_iteration = x[y.index(min_energy)]
        label = f"Minimum Value: {min_energy}\n"

        ax.text(
            min_energy_iteration, min_energy + 0.1 * (max(y) - min(y)), label, color="r"
        )
        ax.set_ylabel("Value")
        ax.set_xlabel("Iteration")
        ax.set_title("Optimization Result")
        ax.legend()

        if ax is None:
            plt.show()


class OptimizationJob(Job[OptimizationResult]):

    def results(self) -> OptimizationResult:
        if self.id is None:
            raise ValueError("Optimization Job not yet run")
        results = OptimizationResultSchema.parse_obj(self.backend.results(self.id))
        progress = self.backend.get_job_progress(self.id)

        return OptimizationResult(
            minimum_value=results.solution.minimum_value,
            optimal_parameters=results.solution.optimal_parameters,
            progress=progress,
        )


class Optimization(Workload[OptimizationJob]):
    def __init__(
        self,
        quantum_function: QuantumFunction,
        method: str,
        initial_params: Optional[List[Union[float, int]]] = None,
        log_interval: int = 1,
        options: Optional[dict] = None,
        maxiter: Optional[int] = None,
        graph: Optional[nx.Graph] = None,
        variables: Optional[List[str]] = None,
        quadratic_expression: Optional[str] = None,
    ):
        self.quantum_function = quantum_function
        self.method = method
        self.initial_params = initial_params
        self.log_interval = log_interval
        self.options = options
        self.maxiter = maxiter
        self.graph = graph
        self.variables = variables
        self.quadratic_expression = quadratic_expression

    def to_workload_input(
        self,
        params: Optional[List[float]] = None,
    ) -> WorkloadInput:
        qf_input = self.quantum_function.to_workload_input().input.data
        return WorkloadInput(
            type=JobType.optimization,
            input=OptimizationWorkload(
                quantum_function=qf_input,
                method=OptimizationMethod(self.method),
                initial_params=self.initial_params,
                log_interval=self.log_interval,
                options=self.options,
                maxiter=self.maxiter,
            ),
        )

    def create_job(self, id: str, job: JobRequest, backend: Backend) -> OptimizationJob:
        return OptimizationJob(id, job, backend)



from typing import Optional, List, Dict, Any, Type

from ionq.schemas.job import (
    JobRequest,
    TargetOptions,
    JobProgress,
    NoiseOptions,
    NoiseModel,
)
from ionq import Client

from .base import JobType, Workload
from .circuit import CircuitJob
from .quantum_function import QuantumFunctionJob
from .optimization import OptimizationJob


class Backend:
    def __init__(
        self, name: str, client: Optional[Client] = None, gateset: str = "qis"
    ):
        self.name = name
        self.client = client if client is not None else Client()
        self.gateset = gateset

    def run(
        self,
        workload: Workload[JobType],
        name: Optional[str] = None,
        shots: Optional[int] = None,
        params: Optional[List[float]] = None,
        noise: Optional[str] = None,
    ) -> JobType:

        workload_input = workload.to_workload_input(params)
        request = JobRequest(
            name=name,
            backend=TargetOptions(self.name),
            type=workload_input.type,
            input=workload_input.input,
            shots=shots,
            noise=NoiseOptions(model=NoiseModel(noise)) if noise else None,
        )

        job = self.client.jobs.create_job(request)

        return workload.create_job(job.id, request, self)

    def get_job_progress(self, job_id: str) -> JobProgress:
        return self.client.jobs.get_job_progress(job_id)

    def get_optimization_job(self, job_id: str) -> OptimizationJob:
        return self.get_job(job_id, OptimizationJob)

    def get_circuit_job(self, job_id: str) -> CircuitJob:
        return self.get_job(job_id, CircuitJob)

    def get_quantum_function_job(self, job_id: str) -> QuantumFunctionJob:
        return self.get_job(job_id, QuantumFunctionJob)

    def get_job(self, job_id: str, job_type: Type[JobType]) -> JobType:
        job = self.client.jobs.get_job(job_id)
        return job_type(job_id, job, self)

    def get_job_results(self, uri: str) -> Dict[str, Any]:
        return self.client.jobs.get_job_results(uri)

    def results(
        self,
        job_id: str,
    ) -> Dict[str, Any]:
        return self.client.jobs.wait_for_results(job_id)



class IonQAPIError(Exception):
    """Base class for all API errors."""

    def __init__(self, error=None):
        super().__init__(error or "An unspecified error occurred.")


class UserCancelledError(Exception):
    """Exception raised when the user cancels an operation."""

    def __init__(self, error=None):
        super().__init__(error or "The user cancelled the operation.")


class PreflightError(IonQAPIError):
    """Exception raised for errors that occur before execution starts."""

    pass


class InternalError(IonQAPIError):
    """Exception raised for internal errors within the API."""

    pass


class TooManyGates(IonQAPIError):
    """Exception raised when too many gates are requested."""

    pass


class SystemCancel(IonQAPIError):
    """Exception raised when a system administrator cancels a process."""

    pass


class QuotaExhaustedError(IonQAPIError):
    """Exception raised when the quota for resources has been exhausted."""

    pass


class TimedOut(IonQAPIError):
    """Exception raised when the operation times out."""

    pass


class TooLongPredictedExecutionTime(IonQAPIError):
    """Exception raised when the predicted execution time is too long."""

    pass


class CircuitTooLarge(IonQAPIError):
    """Exception raised when the circuit size exceeds the limit."""

    pass


class TooManyShots(IonQAPIError):
    """Exception raised when the number of shots exceeds the limit."""

    pass


class ContractExpiredError(IonQAPIError):
    """Exception raised when the contract has expired."""

    pass


class SimulationTimeout(IonQAPIError):
    """Exception raised when a simulation times out."""

    pass


class OptimizationError(IonQAPIError):
    """Exception raised during optimization process errors."""

    pass


class UnsupportedGate(IonQAPIError):
    """Exception raised for using unsupported gates."""

    pass


class CompilationError(IonQAPIError):
    """Exception raised during compilation errors."""

    pass


class SimulationError(IonQAPIError):
    """Exception raised during simulation errors."""

    pass


class Invalid(IonQAPIError):
    """Exception raised for invalid operations."""

    pass


class BillingError(IonQAPIError):
    """Exception raised for billing errors."""

    pass


class TooManyControls(IonQAPIError):
    """Exception raised when too many control gates are used."""

    pass


class DebiasingError(IonQAPIError):
    """Exception raised during debiasing process errors."""

    pass


class Failed(IonQAPIError):
    """Exception raised when a process fails without a specific category."""

    pass


class QuantumComputerError(IonQAPIError):
    """Exception raised for errors specific to quantum computer operations."""

    pass


class Expired(IonQAPIError):
    """Exception raised when a session or token has expired."""

    pass


class ExecutionError(IonQAPIError):
    """Exception raised for execution failures."""

    pass


class NotEnoughQubits(IonQAPIError):
    """Exception raised when there are not enough qubits available."""

    pass



from .circuit import Circuit
from .quantum_function import HamiltonianEnergy
from .optimization import Optimization, OptimizationMethod

__all__ = ["Circuit", "HamiltonianEnergy", "Optimization", "OptimizationMethod"]



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Optional,
    List,
)

import matplotlib.pyplot as plt

from ionq.schemas.job import (
    JobRequest,
    WorkloadInput,
    QuantumFunctionInput,
    HamiltonianEnergyQuantumFunction,
    HamiltonianEnergyData,
    Hamiltonian,
    Ansatz,
    JobType
)

from .base import Result, Job, Workload

if TYPE_CHECKING:
    from .backend import Backend


class QuantumFunctionResult(Result):
    def __init__(self, value: float, variance: float):
        self.value = value
        self.variance = variance

    def plot_results(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 6))

        ax.errorbar(
            ["Value"],
            [self.value],
            yerr=[self.variance],
            fmt="o",
            capsize=10,
        )
        ax.set_ylabel("Value")
        ax.set_title("Quantum Function Result")

        if ax is None:
            plt.show()


class QuantumFunctionJob(Job[QuantumFunctionResult]):

    def results(self) -> QuantumFunctionResult:
        if self.id is None:
            raise ValueError("Quantum Function not yet run")
        results_data = self.backend.results(self.id)

        return QuantumFunctionResult(
            value=results_data["value"],
            variance=results_data["variance"],
        )


class QuantumFunction(Workload[QuantumFunctionJob]):
    pass


class HamiltonianEnergy(QuantumFunction):
    def __init__(
        self,
        ansatz: Ansatz,
        hamiltonian: Hamiltonian,
    ):
        assert (
            ansatz.num_qubits == hamiltonian.num_qubits
        ), "Ansatz and Hamiltonian must have the same number of qubits"
        self.ansatz = ansatz
        self.hamiltonian = hamiltonian

    def to_workload_input(
        self,
        params: Optional[List[float]] = None,
    ) -> WorkloadInput:
        return WorkloadInput(
            type=JobType.quantum_function,
            input=QuantumFunctionInput(
                data=HamiltonianEnergyQuantumFunction(
                    data=HamiltonianEnergyData(
                        hamiltonian=self.hamiltonian.terms,
                        ansatz=self.ansatz,
                    ),
                ),
                params=params,
            ),
        )

    def create_job(
        self, id: str, job: JobRequest, backend: Backend
    ) -> QuantumFunctionJob:
        return QuantumFunctionJob(id, job, backend)



from typing import List, Union, Optional
import numpy as np
from dataclasses import dataclass
from pydantic import BaseModel, Field

from qiskit import QuantumCircuit
from qiskit.quantum_info import SparsePauliOp
from tqdm import tqdm
from ionq import Backend
from ionq.jobs.quantum_function import QuantumFunction, QuantumFunctionJob
from ionq.schemas.job import (
    Ansatz,
    JobType,
    WorkloadInput,
    QuantumFunctionInput,
    HamiltonianEnergyQuantumFunction,
    JobRequest,
    HamiltonianPauliTerm,
)
from ionq.utils import ionq_qiskit
from .problem import Problem


@dataclass
class OptimalSolution:
    x: str
    fun: float


class LinearConstraint(BaseModel):
    r"""
    A class to model linear inequality constraints of the form

    .. math::

        a^T x \leq b.

    INPUT:

        - ``coeffs`` -- a list of floats describing the coefficient vector `a`
        - ``rhs`` -- the Right-Hand-Side coefficient `b`
    """

    coeffs: List[float]
    rhs: float


class QuadraticConstraint(BaseModel):
    r"""
    A class to model quadratic inequality constraints of the form

    .. math::

        x^T P x + r^T x \leq c.

    INPUT:

        - ``quadratic_coeff`` -- a list of list of floats describing the
          symmetric matrix `P`
        - ``linear_coeff`` -- a list of floats describing the row vector `r`
        - ``rhs`` -- the Right-Hand-Side coefficient `c`
    """

    quadratic_coeff: List[List[float]]
    linear_coeff: List[float]
    rhs: float


class MatrixLinearConstraint:
    def __init__(self, constraints: List[LinearConstraint]):
        self.linear_coeff = np.array([lc.coeffs for lc in constraints])
        self.rhs_vec = np.array([lc.rhs for lc in constraints])
        self.num_constraints = len(constraints)

    def __iter__(self):
        for coeffs, rhs in zip(self.linear_coeff, self.rhs_vec):
            yield LinearConstraint(coeffs=coeffs.tolist(), rhs=rhs.tolist())

    def compute_violations(self, x: np.ndarray) -> np.ndarray:
        """
        Compute the violations for this matrix constraint.
        """
        if self.num_constraints == 0:
            return np.array([])

        lhs = self.linear_coeff @ x - self.rhs_vec
        vv = np.zeros_like(lhs)
        vv[lhs > 0] = lhs[lhs > 0]
        return vv


class MatrixQuadraticConstraint:
    def __init__(self, constraints: List[QuadraticConstraint]):
        self.quadratic_coeff = np.array([lc.quadratic_coeff for lc in constraints])
        self.linear_coeff = np.array([lc.linear_coeff for lc in constraints])
        self.rhs_vec = np.array([lc.rhs for lc in constraints])
        self.num_constraints = len(constraints)

    def __iter__(self):
        for quad_coeff, lin_coeff, rhs in zip(
            self.quadratic_coeff, self.linear_coeff, self.rhs_vec
        ):
            yield QuadraticConstraint(
                quadratic_coeff=quad_coeff.tolist(),
                linear_coeff=lin_coeff.tolist(),
                rhs=rhs.tolist(),
            )

    def compute_violations(self, x: np.ndarray) -> np.ndarray:
        """
        Compute the violations for this matrix constraint.
        """
        if self.num_constraints == 0:
            return np.array([])

        lhs = x.T @ self.quadratic_coeff @ x + self.linear_coeff @ x - self.rhs_vec
        vv = np.zeros_like(lhs)
        vv[lhs > 0] = lhs[lhs > 0]
        return vv


class QCQPHamiltonianData(BaseModel):
    hamiltonian: List[HamiltonianPauliTerm]
    ansatz: Ansatz
    linear_constraints: Optional[List[LinearConstraint]] = None
    quadratic_constraints: Optional[List[QuadraticConstraint]] = None
    penalty: Optional[float] = None

    def __init__(self, **data):
        super().__init__(**data)
        if self.penalty is not None and (
            self.linear_constraints is None and self.quadratic_constraints is None
        ):
            raise ValueError("Constraints must be set if penalty is set")


class QCQPHamiltonianFunction(HamiltonianEnergyQuantumFunction):
    type: str = Field(default="hamiltonian-energy", const=True)
    data: QCQPHamiltonianData


class QCQPObjective(QuantumFunction):
    def __init__(
        self,
        ansatz: Ansatz,
        qp_objective: np.ndarray,
        linear_constraints: List[LinearConstraint] = [],
        quadratic_constraints: List[QuadraticConstraint] = [],
        penalty: Optional[float] = None,
    ):

        self.ansatz = ansatz
        self._qp_objective = qp_objective
        self.linear_constraints = linear_constraints
        self.quadratic_constraints = quadratic_constraints
        self.penalty = penalty

    def get_qubo_objective_as_ising_ham(self) -> SparsePauliOp:
        r"""
        Compute the Ising Hamiltonian obtained by applying the Ising map

        $$x_j \rightarrow \frac{1}{2}(I - \sigma_j^Z),$$

        with $\sigma_j^Z$ denoting the Pauli-$Z$ gate acting on the $j$th
        qubit, to the QUBO objective $x^T Q x$ specified by the symmetric
        2D-NumPy array ``self.qp_objective``.
        """
        if not hasattr(self, "_ising_ham"):
            Q = self._qp_objective
            terms = [("I" * Q.shape[0], 1 / 4 * (Q.sum() + Q.trace()))]

            # Linear terms
            for j, sj in enumerate(
                filter(lambda v: not np.isclose(v, 0), Q.sum(axis=1))
            ):
                lin = "".join(
                    "I" if k != j else "Z" for k in reversed(range(Q.shape[0]))
                )
                terms += [(lin, -1 / 2 * sj)]

            # Quadratic terms
            for i in range(Q.shape[0]):
                for j in range(i + 1, Q.shape[0]):
                    if not np.isclose(Q[i, j], 0):
                        quad = "".join(
                            "I" if k not in [i, j] else "Z"
                            for k in reversed(range(Q.shape[0]))
                        )
                        terms += [(quad, 1 / 2 * Q[i, j])]
            self._ising_ham = SparsePauliOp.from_list(terms)
        return self._ising_ham

    def to_workload_input(
        self,
        params: Optional[List[float]] = None,
    ) -> WorkloadInput:
        ising_ham = self.get_qubo_objective_as_ising_ham()
        assert (
            self.ansatz.num_qubits == ising_ham.num_qubits
        ), "Ansatz and Hamiltonian must have the same number of qubits"
        return WorkloadInput(
            type=JobType.quantum_function,
            input=QuantumFunctionInput(
                data=QCQPHamiltonianFunction(
                    data=QCQPHamiltonianData(
                        ansatz=self.ansatz,
                        hamiltonian=ionq_qiskit.to_hamiltonian(ising_ham).terms,
                        linear_constraints=self.linear_constraints or None,
                        quadratic_constraints=self.quadratic_constraints or None,
                        penalty=self.penalty,
                    ),
                ),
                params=params,
            ),
        )

    def create_job(
        self, id: str, job: JobRequest, backend: Backend
    ) -> QuantumFunctionJob:
        return QuantumFunctionJob(id, job, backend)


class QuadraticProgram(Problem):
    r"""
    A class to solve Quadratically-Constrained Quadratic Programs (QCQPs).

    The ``objective`` $x^T Q x$ may be specified by a symmetric 2D-NumPy array
    (whose diagonal defines objective's linear component) or as the 
    objective of a Qiskit ``QuadraticProgram`` instance.

    Optionally, a list of linear and/or quadratic ``constraints`` may be
    supplied. Each element of ``constraints`` is either a pair ``(A, b)`` or a
    triple ``(P, r, c)``: 

    - if ``len(constraints[j]) == 2``, then ``constraints[j]`` specifies the
      linear inequality $Ax \leq b$; 
    - otherwise, ``constraints[j]`` defines the quadratic inequality
      $x^T P x + r^T x \leq c$.

    Additionally, a ``penalty`` parameter may be provided, which determines the
    magnitude of the penalty terms added to the Ising energies corresponding to
    infeasible states.

    In symbols, this class models the following QCQP:

    $$
    \begin{aligned}
    \min_x \quad & x^T Q x \\
    \text{s.t.} \quad & A_j x \leq b_j, \\
    & x^T P_j x + r_j^T x \leq c, \\
    & x_k \in \{0, 1\}, \quad \forall k.
    \end{aligned}
    $$
    """

    def __init__(
        self,
        objective: np.ndarray,
        linear_constraints: List[LinearConstraint] = [],
        quadratic_constraints: List[QuadraticConstraint] = [],
    ):
        # Standardize objective
        self._qp_objective = objective

        # Initialize constraints handlers
        self._mat_lc = MatrixLinearConstraint(linear_constraints)
        self._mat_qc = MatrixQuadraticConstraint(quadratic_constraints)

    def compute_optimal_solution(self):
        """
        Compute the optimal (feasible) solution of ``self``.

        Currently this method finds the global optimal solution by enumeration.
        """
        if hasattr(self, "_optimal_solution"):
            return self._optimal_solution

        # Find the feasible states
        n = self._qp_objective.shape[0]
        best_soln = (0, 1e100)
        base = np.arange(n, dtype=int)[np.newaxis, ::-1]
        chunk_sz = min(2**n, 2**20)

        # Semi-vectorized loop: iterate over chunks of 2**20 states at a time
        print("Finding optimal solution one batch of states at a time...")
        for chunk_idx in tqdm(np.arange(2 ** max(n - 20, 0))):
            chunk = np.arange(chunk_idx * chunk_sz, (chunk_idx + 1) * chunk_sz)[
                :, np.newaxis
            ]
            states = np.array(2**base & chunk > 0, dtype=int)
            feasible_idx = np.zeros(len(states), dtype=bool)
            for j in np.arange(len(states)):
                feasible_idx[j] = self.is_feasible(states[j])
            X = states[feasible_idx]
            obj_vals = (X * np.dot(X, self._qp_objective)).sum(axis=1)
            chunk_min_idx = obj_vals.argmin()
            if obj_vals[chunk_min_idx] < best_soln[1]:
                best_soln = (X[chunk_min_idx], obj_vals[chunk_min_idx])
        self._optimal_solution = OptimalSolution(
            "".join(map(str, best_soln[0])), best_soln[1]
        )
        return self._optimal_solution

    def compute_violations(self, x: Union[np.ndarray, str]):
        r"""
        Compute the list constraint violations, defined as the positive part of
        $A_j x - b_j$ for every linear inequality and $x^T P x + r_j^T x - c$
        for every quadratic one.
        """
        if isinstance(x, str):
            x = np.array(list(x), dtype=int)

        lc_viol = self._mat_lc.compute_violations(x)
        qc_viol = self._mat_qc.compute_violations(x)
        return np.hstack((lc_viol, qc_viol))

    def evaluate_qp_objective(self, x: Union[np.ndarray, str]) -> float:
        """
        Evaluate the QP objective at the given bit-string ``x``, which may be
        given as a NumPy array or a string.
        """
        if isinstance(x, str):
            x = np.array(list(x), dtype=int)
        return x.T @ self.qp_objective @ x

    @classmethod
    def from_docplex_quadratic_program(cls, docplex_qp):
        """
        Construct a :class:`.QuadraticProgram` instance from a `docplex` QP.
        """
        if docplex_qp.quadratic_constraints:
            raise NotImplementedError(
                "Only linear constraints are currently supported..."
            )

        # Interpret linear constraints
        constraints = list()
        for lc in docplex_qp.linear_constraints:
            ionq_lc = LinearConstraint(coeffs=lc.linear.to_array(), rhs=lc.rhs)

            # Check the type of constraint
            if lc.sense.name == "EQ":
                constraints.append(
                    LinearConstraint(coeffs=-lc.linear.to_array(), rhs=-lc.rhs)
                )
            else:
                print(f"missing constraint with sense {lc.sense.name}")
            constraints.append(ionq_lc)
        return QuadraticProgram(docplex_qp.objective, constraints)

    def is_feasible(self, x: Union[np.ndarray, str]) -> bool:
        """
        Determine if the bit-string ``x`` satisfies the problem constraints.
        """
        if isinstance(x, str):
            x = np.array(list(x), dtype=int)
        return np.allclose(self.compute_violations(x), 0)

    @property
    def linear_constraints(self) -> List[LinearConstraint]:
        """
        Get the constraints of ``self``, if any.
        """
        return list(self._mat_lc)

    @property
    def qp_objective(self) -> np.ndarray:
        """
        Get the symmetric 2D-array describing the objective of ``self``.
        """
        return self._qp_objective

    @property
    def quadratic_constraints(self) -> List[QuadraticConstraint]:
        """
        Get the constraints of ``self``, if any.
        """
        return list(self._mat_qc)

    def quantum_function_objective(
        self, ansatz: QuantumCircuit, penalty: float = 10
    ) -> QCQPObjective:
        """
        Get the :class:`.QuantumFunction` describing `self`'s optimization
        objective.
        """
        return QCQPObjective(
            ansatz=ionq_qiskit.to_ansatz(ansatz),
            qp_objective=self._qp_objective,
            linear_constraints=self.linear_constraints,
            quadratic_constraints=self.quadratic_constraints,
            penalty=penalty,
        )



from typing import Dict, Any, Callable, Optional
import json
import gzip
import base64


def compress(obj) -> str:
    return base64.b64encode(gzip.compress(json.dumps(obj).encode("utf-8"))).decode(
        "utf-8"
    )


class Circuit:
    def __init__(
        self,
        n_qubits: int,
        shots: int = 100,
        format: str = "ionq.circuit.v0",
        gateset: str = "qis",
        registers: Optional[list] = None,
        name: Optional[str] = None,
        error_mitigation: Optional[Dict[str, bool]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        noise: Optional[Dict[str, str]] = None,
        target: str = "simulator",
    ):
        self.n_qubits = n_qubits
        self.shots = shots
        self.name = name
        self.error_mitigation = error_mitigation
        self.metadata = metadata
        self.noise = noise
        self.target = target
        self.input = {
            "format": format,
            "gateset": gateset,
            "qubits": n_qubits,
            "circuit": [],
        }
        self.circuit = self.input["circuit"]
        self.registers = registers

    def create_job_payload(self, *additional_circuits) -> dict:
        all_circuits = [self] + list(*additional_circuits)

        job_payload = {
            "shots": self.shots,
            "target": self.target,
            "noise": (
                {
                    "model": self.noise["model"] if self.noise else None,
                    "seed": (
                        self.noise["seed"]
                        if self.noise and "seed" in self.noise
                        else None
                    ),
                }
                if self.noise
                else None
            ),
            "error_mitigation": self.error_mitigation,
            "input": {
                "format": self.input["format"],
                "gateset": self.input["gateset"],
            },
        }

        job_payload["name"] = (
            ", ".join([circuit.name for circuit in all_circuits if circuit.name])
            or None
        )
        job_payload["metadata"] = {
            "compressed": compress(
                [circuit.metadata for circuit in all_circuits if circuit.metadata]
            )
        }
        job_payload["input"]["circuits"] = [
            {
                "qubits": circuit.n_qubits,
                "registers": circuit.registers,
                "circuit": circuit.circuit,
            }
            for circuit in all_circuits
        ]

        def remove_none(obj):  # https://stackoverflow.com/a/20558778
            if isinstance(obj, (list, tuple, set)):
                return type(obj)(remove_none(x) for x in obj if x is not None)
            elif isinstance(obj, dict):
                return type(obj)(
                    (remove_none(k), remove_none(v))
                    for k, v in obj.items()
                    if k is not None and v is not None
                )
            else:
                return obj

        job_payload = remove_none(job_payload)

        return job_payload

    def __bool__(self) -> bool:
        return self.n_qubits > 0 and bool(self.circuit)

    def __getattr__(self, name: str) -> Callable[..., Any]:
        def method(*args, **kwargs):
            self.append(name, *args, **kwargs)

        return method

    def append(self, name: str, *args, **kwargs):
        if not name:
            raise ValueError("No gate name provided")

        gate = {"gate": name, "targets": [], "controls": []}

        # Handling positional arguments for target and control qubits
        if args:
            # Single qubit target or first qubit as control in case of multiple args
            (
                gate["targets"].append(args[0])
                if len(args) == 1
                else gate["controls"].append(args[0])
            )
            if len(args) > 1:  # Additional targets if more than one arg
                gate["targets"].extend(args[1:])

        # Handling keyword arguments for gate configuration
        for key, value in kwargs.items():
            if key in ("control", "controls", "target", "targets", "rotation"):
                if isinstance(value, list):
                    gate[key + "s" if key[-1] != "s" else key] = value
                else:
                    gate[
                        (
                            key + "s"
                            if key[-1] != "s" and key not in ("rotation", "controls")
                            else key
                        )
                    ] = ([value] if key in ("target", "control") else value)
            else:
                gate[key] = value

        self.circuit.append(gate)



from typing import Any, Dict, Optional
import os
import time

import requests
from requests.exceptions import ConnectionError, Timeout

from ..errors import IonQAPIError
from ..defaults import DEFAULT_API_URL, SDK_VERSION
from . import resources
from .. import errors


def handle_error(response_json: dict):
    if "failure" in response_json and response_json["failure"] is not None:
        error_code = response_json["failure"]["code"]
        error = response_json["failure"].get("error") or response_json["failure"].get(
            "message"
        )
        if error_code in code_to_class:
            # Raise the appropriate error class
            raise code_to_class[error_code](error=error)
        else:
            raise errors.IonQAPIError(error=error)


code_to_class = {
    "PreflightError": errors.PreflightError,
    "InternalError": errors.InternalError,
    "TooManyGates": errors.TooManyGates,
    "SystemCancel": errors.SystemCancel,
    "QuotaExhaustedError": errors.QuotaExhaustedError,
    "TimedOut": errors.TimedOut,
    "TooLongPredictedExecutionTime": errors.TooLongPredictedExecutionTime,
    "CircuitTooLarge": errors.CircuitTooLarge,
    "TooManyShots": errors.TooManyShots,
    "ContractExpiredError": errors.ContractExpiredError,
    "SimulationTimeout": errors.SimulationTimeout,
    "OptimizationError": errors.OptimizationError,
    "UnsupportedGate": errors.UnsupportedGate,
    "CompilationError": errors.CompilationError,
    "SimulationError": errors.SimulationError,
    "Invalid": errors.Invalid,
    "BillingError": errors.BillingError,
    "TooManyControls": errors.TooManyControls,
    "DebiasingError": errors.DebiasingError,
    "Failed": errors.Failed,
    "QuantumComputerError": errors.QuantumComputerError,
    "Expired": errors.Expired,
    "ExecutionError": errors.ExecutionError,
    "NotEnoughQubits": errors.NotEnoughQubits,
}


class Client:

    jobs: resources.JobsResource
    backends: resources.BackendsResource

    api_key: str
    base_url: str

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
    ) -> None:

        if api_key is None:
            api_key = os.environ.get("IONQ_API_KEY")
        if api_key is None:
            raise IonQAPIError(
                "The api_key client option must be set either by passing api_key to the client or by setting the IONQ_API_KEY environment variable"
            )
        self.api_key = api_key

        if base_url is None:
            base_url = os.environ.get("IONQ_BASE_URL") or DEFAULT_API_URL
        self.base_url = base_url

        self.jobs = resources.JobsResource(self)
        self.backends = resources.BackendsResource(client=self)

    @property
    def headers(self) -> dict[str, str]:
        return {
            "Authorization": f"apiKey {self.api_key}",
            "Content-Type": "application/json",
            "User-Agent": f"ionq/{SDK_VERSION}",
        }

    def request(
        self,
        url: str,
        method: str,
        max_retries: int = 3,
        timeout_seconds: int = 10,
        sleep_seconds: int = 2,
        **kwargs,
    ) -> Dict[str, Any]:
        retries = 0
        while retries < max_retries:
            try:
                response = requests.request(
                    method,
                    f"{self.base_url}{url}",
                    timeout=timeout_seconds,
                    headers=self.headers,
                    **kwargs,
                )
                response.raise_for_status()  # Raises HTTPError for bad HTTP responses
                response_json: Dict[str, Any] = response.json()  # Raises ValueError
                handle_error(response_json)  # Raises API-specific errors
                return response_json
            except (ConnectionError, Timeout):
                if retries == max_retries - 1:
                    raise  # Re-raise the last exception if out of retries
                retries += 1
                time.sleep(sleep_seconds * (2 ** (retries - 1)))  # Exponential backoff
            except ValueError:
                # Handling JSON parsing failure
                print(f"Failed to parse JSON response: {response.text}")
                raise
            except requests.exceptions.RequestException as e:
                print(f"Request failed: {e}")
                raise  # For non-retriable exceptions, we raise them immediately
        raise Timeout(f"Request timed out after {max_retries} retries")



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Generic,
    TypeVar,
    Optional,
    List,
)


from ionq.schemas.job import (
    JobRequest,
    JobDetails,
    WorkloadInput,
)


if TYPE_CHECKING:
    from .backend import Backend

R = TypeVar("R", bound="Result")
JobType = TypeVar("JobType", bound="Job")


class Job(ABC, Generic[R]):

    details: JobRequest | JobDetails

    def __init__(self, id: str, details: JobRequest | JobDetails, backend: Backend):
        self.id = id
        self.details = details
        self.backend = backend

    @abstractmethod
    def results(self) -> R:
        pass


class Workload(ABC, Generic[JobType]):
    @abstractmethod
    def to_workload_input(
        self,
        params: Optional[List[float]] = None,
    ) -> WorkloadInput:
        pass

    @abstractmethod
    def create_job(self, id: str, job: JobRequest, backend: Backend) -> JobType:
        pass


class Result(ABC):
    @abstractmethod
    def plot_results(self, ax=None):
        pass



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Optional,
    List,
    Dict,
    Literal,
    Any,
)


import matplotlib.pyplot as plt

from ionq.schemas.job import (
    JobRequest,
    MultiCircuitWorkload,
    WorkloadInput,
    JobType,
    CircuitWorkload,
    GateSet,
)

from .base import Result, Job, Workload

if TYPE_CHECKING:
    from .backend import Backend


class CircuitResult(Result):
    def __init__(
        self, probabilities: Dict[str, float], counts: Optional[Dict[str, int]] = None
    ):
        self.counts = counts
        self.probabilities = probabilities

    def top_candidates(self, n: Optional[int] = None) -> List[str]:
        if n is None:
            n = len(self.probabilities)
        return sorted(self.probabilities, key=self.probabilities.get, reverse=True)[:n]  # type: ignore

    def plot_results(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 6))

        ax.bar(list(self.probabilities.keys()), list(self.probabilities.values()))
        ax.set_ylabel("Probability")
        ax.set_xlabel("State")
        ax.set_title("Circuit Result")

        if ax is None:
            plt.show()


class CircuitJob(Job[CircuitResult]):

    def _get_counts(self, probabilities: Dict[str, float]) -> Optional[Dict[str, int]]:
        if self.details.shots is None:
            return None
        return {k: int(v * self.details.shots) for k, v in probabilities.items()}

    def results(self) -> List[CircuitResult]:
        if self.id is None:
            raise ValueError("Circuit Job not yet run")

        results = self.backend.results(self.id)

        if self.details.type == JobType.multi_circuit:
            all_probabilities = self.backend.get_job_results(results.get("probabilities", {}).get("url"))
            return [
                CircuitResult(
                    probabilities=probabilities,
                    counts=self._get_counts(probabilities),
                )
                for probabilities in all_probabilities.values()
            ]
        else:
            if "histogram" in results:
                counts = self.backend.get_job_results(results["histogram"]["url"])
                probabilities = {
                    k: v / self.details.shots for k, v in counts.items()
                } if self.details.shots else {}
                return [
                    CircuitResult(
                        probabilities=probabilities,
                        counts=counts,
                    )
                ]
            probabilities = self.backend.get_job_results(results.get("probabilities", {}).get("url"))
            return [
                CircuitResult(
                    probabilities=probabilities,
                    counts=self._get_counts(probabilities),
                )
            ]


class Circuit(Workload[CircuitJob]):

    gateset: Literal["qis", "native"]

    def __init__(
        self,
        circuits: List[Dict[str, Any]],
        gateset: Literal["qis", "native"] = "qis",
        qubits: Optional[int] = None,
        name: Optional[str] = None,
    ):
        self.circuits = circuits
        self.gateset = gateset
        self.qubits = qubits
        self.name = name

    def to_workload_input(
        self,
        params: Optional[List[float]] = None,
    ) -> WorkloadInput:
        if len(self.circuits) > 1:
            return WorkloadInput(
                type=JobType.multi_circuit,
                input=MultiCircuitWorkload(
                    gateset=GateSet(self.gateset),
                    qubits=self.qubits,
                    circuits=self.circuits,
                ),
            )
        return WorkloadInput(
            type=JobType.circuit,
            input=CircuitWorkload(
                gateset=GateSet(self.gateset),
                qubits=self.qubits,
                circuit=self.circuits[0]["circuit"],
            ),
        )

    def create_job(self, id: str, request: JobRequest, backend: Backend) -> CircuitJob:
        return CircuitJob(id, request, backend)



from typing import Optional
from abc import ABC, abstractmethod

from qiskit import QuantumCircuit
from ionq.jobs.quantum_function import QuantumFunction


class Problem(ABC):
    """
    An abstract class to model an optimization problem.

    The constructor, defined in a child class, ingests problem instance data.
    Then the class's main method, :meth:`.quantum_function_objective`
    constructs the problem's objective function as a :class:`.QuantumFunction`
    object that can be passed to a hybrid solver for optimization.
    """

    @abstractmethod
    def quantum_function_objective(
        self,
        ansatz: Optional[QuantumCircuit] = None,
        **kwargs,
    ) -> QuantumFunction:
        pass



from .backend import Backend
from .characterization import Characterization
from .circuit import CircuitOperation
from .job import (
    JobRequest,
    JobDetails,
    JobStatus,
    JobUpdate,
    JobProgress,
)


__all__ = [
    "JobRequest",
    "JobDetails",
    "JobStatus",
    "JobUpdate",
    "Backend",
    "Characterization",
    "CircuitOperation",
    "JobProgress",
]



import numpy as np
import sympy as sp
import random
import json
import networkx as nx
from qiskit.quantum_info import SparsePauliOp


def generate_random_quadratic_equation(
    num_variables, coeff_range=(-3, 3), constant_range=(-3, 3)
):
    # Create symbolic variables x1, x2, ..., xn
    variables = [sp.Symbol(f"x{i+1}") for i in range(num_variables)]

    # Generate random coefficients for the linear and quadratic terms
    linear_terms = [random.randint(*coeff_range) * var for var in variables]
    quadratic_terms = [
        random.randint(*coeff_range) * var1 * var2
        for i, var1 in enumerate(variables)
        for var2 in variables[i:]
    ]

    # Generate random constant term
    constant = random.randint(*constant_range)

    # Construct the quadratic equation
    quadratic_expr = sum(linear_terms) + sum(quadratic_terms)
    equation = sp.Eq(quadratic_expr, constant)

    return f"{sp.simplify(equation.lhs)} = {equation.rhs}"


def graph_to_qubo(graph: nx.Graph) -> np.ndarray:
    num_nodes = graph.number_of_nodes()
    Q = np.zeros((num_nodes, num_nodes))

    for u, v, data in graph.edges(data=True):
        weight = data.get("weight", 1)  # Assume a default weight of 1 if not specified
        Q[u, v] -= weight
        Q[v, u] -= weight
        Q[u, u] += weight
        Q[v, v] += weight

    return Q


def qubo_obj_to_ising_ham(Q: np.ndarray):
    num_qubits = Q.shape[0]
    hamiltonian = [("I" * num_qubits, 1 / 4 * (Q.sum() + Q.trace()))]

    # Linear terms
    for j, sj in enumerate(filter(lambda v: not np.isclose(v, 0), Q.sum(axis=1))):
        lin = "".join("I" if k != j else "Z" for k in reversed(range(num_qubits)))
        hamiltonian += [(lin, -1 / 2 * sj)]

    # Quadratic terms
    for i in range(num_qubits):
        for j in range(i + 1, num_qubits):
            if not np.isclose(Q[i, j], 0):
                quad = "".join(
                    "I" if k not in [i, j] else "Z" for k in reversed(range(num_qubits))
                )
                hamiltonian += [(quad, 1 / 2 * Q[i, j])]
    return SparsePauliOp.from_list(hamiltonian)


class SafeEncoder(json.JSONEncoder):
    """Custom JSON encoder to handle non-JSON-safe objects."""

    def default(self, o):
        try:
            return super().default(o)
        except Exception:
            return str(o)



from typing import List, Tuple
from pydantic import BaseModel


class FidelityDetail(BaseModel):
    mean: float


class Fidelity(BaseModel):
    _1q: FidelityDetail
    _2q: FidelityDetail
    spam: FidelityDetail


class Timing(BaseModel):
    readout: float
    reset: float
    t1: float
    t2: float
    _1q: float
    _2q: float


class Characterization(BaseModel):
    id: str
    date: int
    backend: str
    qubits: int
    connectivity: List[Tuple[int, int]]
    fidelity: Fidelity
    timing: Timing

    def __repr__(self):
        # Create a summary of the connectivity showing the first two and the last connection
        if len(self.connectivity) > 2:
            connectivity_summary = f"{self.connectivity[0]}, {self.connectivity[1]}, ..., {self.connectivity[-1]}"
        else:
            connectivity_summary = ", ".join(str(pair) for pair in self.connectivity)

        # Return a formatted string with the summary
        return (
            f"Characterization(id={self.id}, date={self.date}, backend={self.backend}, "
            f"qubits={self.qubits}, connectivity=[{connectivity_summary}], "
            f"fidelity={self.fidelity}, timing={self.timing})"
        )

    class Config:
        schema_extra = {
            "example": {
                "id": "42450d37-33ff-4c92-8d6c-fbe8857e1b5c",
                "date": 1713860166,
                "backend": "qpu.aria-1",
                "qubits": 25,
                "connectivity": [
                    (0, 1),
                    # ...
                    (23, 24),
                ],
                "fidelity": {
                    "1q": {"mean": 0.9997},
                    "2q": {"mean": 0.9706},
                    "spam": {"mean": 0.9948},
                },
                "timing": {
                    "readout": 0.0003,
                    "reset": 0.00002,
                    "t1": 100,
                    "t2": 1,
                    "1q": 0.000135,
                    "2q": 0.0006,
                },
            }
        }



from typing import List, Optional
from pydantic import BaseModel


class CircuitOperation(BaseModel):
    gate: str
    target: Optional[int] = None
    targets: Optional[List[int]] = None
    control: Optional[int] = None
    controls: Optional[List[int]] = None
    rotation: Optional[float] = None


class CircuitList(BaseModel):
    qubits: Optional[int] = None
    registers: Optional[Optional[List[int]]] = None
    circuit: List[CircuitOperation]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..client import Client


class Resource:

    client: Client

    def __init__(
        self,
        client: Client,
    ):
        self.client = client



from pydantic import HttpUrl, BaseModel, validator
from typing import Optional
from enum import Enum
from .characterization import Characterization
from ..defaults import DEFAULT_API_URL


class SystemStatus(str, Enum):
    available = "available"
    unavailable = "unavailable"
    running = "running"
    reserved = "reserved"
    calibrating = "calibrating"
    offline = "offline"
    retired = "retired"


class Backend(BaseModel):
    backend: str
    status: SystemStatus
    average_queue_time: int
    last_updated: int
    degraded: Optional[bool] = None
    characterization_url: Optional[HttpUrl] = None
    characterization: Optional[Characterization] = None

    @validator("characterization_url", pre=True)
    def prepend_url_prefix(cls, v: str):
        return DEFAULT_API_URL + v

    class Config:
        schema_extra = {
            "example": {
                "backend": "qpu.aria-1",
                "status": "available",
                "average_queue_time": 1181215,
                "last_updated": 1490932820,
                "degraded": True,
                "characterization_url": "/characterizations/617a1f8b-59d4-435d-aa33-695433d7155e",
                "characterization": {
                    "id": "3c90c3cc-0d44-4b50-8888-8dd25736052a",
                    "date": 123,
                    "target": "qpu.harmony",
                    "qubits": 1,
                    "connectivity": [[0, 1], [0, 2], [10, 9]],
                    "fidelity": {"spam": {"mean": 1, "stderr": 1}},
                    "timing": {"readout": 123, "reset": 123},
                },
            }
        }



from typing import Optional
import numpy as np
import networkx as nx

from qiskit import QuantumCircuit
from qiskit.circuit import ParameterVector


def BrickworkLayoutAnsatz(
    num_qubits,
    num_layers=2,
    params=None,
    initial_state: Optional[QuantumCircuit] = None,
):
    # Construct a variational ansatz with a brickwork layout structure.
    qc = QuantumCircuit(num_qubits)
    if initial_state:
        qc = initial_state.compose(qc)
        if not qc:
            raise ValueError("Error when adding initial state.")
        qc.barrier()
    else:
        qc.h(range(num_qubits))

    if params is None:
        num_params = np.ceil(num_layers * (num_qubits - 1) / 2).astype(int)
        params = ParameterVector("θ", num_params)
    param_it = iter(params)

    for layer in range(num_layers):
        start_qubit = layer % 2
        for qubit in range(start_qubit, num_qubits - 1, 2):
            qc.cx(qubit, qubit + 1)
            qc.ry(next(param_it), qubit + 1)

    return qc


class GenericQubitEntangler(QuantumCircuit):
    """
    Construct a generic qubit entangler with connectivity specified by a given graph.

    Specify either `num_qubits` or the connectivity `graph`; when using the former,
    we use the complete graph (all-to-all connectivity). By default, the ansatz uses
    one independent parameter per edge in the graph. However, a parameters can be
    specified.
    """

    def __init__(self, num_qubits=None, graph=None, params=None, initial_state=None):
        # Check / standardize input
        if num_qubits is None and graph is None:
            raise ValueError("Must supply at least one of num_qubits or graph")
        if (
            num_qubits is not None
            and graph is not None
            and num_qubits != graph.number_of_nodes()
        ):
            raise ValueError("num_qubits must equal graph.number_of_nodes()!")

        # Impute the number of qubits
        if initial_state is not None:
            num_qubits = initial_state.num_qubits
        else:
            if graph is None:
                graph = nx.complete_graph(num_qubits)
            num_qubits = graph.number_of_nodes()
        super().__init__(num_qubits)

        # Default initial state is |+>
        if initial_state is None:
            self.h(range(self.num_qubits))
        else:
            self.compose(initial_state, inplace=True)
        self.barrier()

        # Construct entangler
        if params is None:
            params = ParameterVector("θ", graph.number_of_edges())
        for theta, (u, v) in zip(params, graph.edges):
            self.cx(u, v)
            self.ry(theta, v)
            self.cx(u, v)



from pydantic import BaseModel, Field
from typing import List, Dict, Optional, Union, Any, Generic, TypeVar
from enum import Enum
from dataclasses import dataclass
import re
from openqasm3 import parser


class JobStatus(str, Enum):
    submitted = "submitted"
    ready = "ready"
    started = "started"
    canceled = "canceled"
    deleted = "deleted"
    completed = "completed"
    failed = "failed"


class NoiseModel(str, Enum):
    ideal = "ideal"
    harmony = "harmony"
    aria_1 = "aria-1"
    forte_1 = "forte-1"


class GateSet(str, Enum):
    qis = "qis"
    native = "native"


class TargetOptions(str, Enum):
    simulator = "simulator"
    harmony = "qpu.harmony"
    aria_1 = "qpu.aria-1"
    aria_2 = "qpu.aria-2"
    forte_1 = "qpu.forte-1"


class NoiseOptions(BaseModel):
    model: Optional[NoiseModel] = None
    seed: Optional[int] = None


class ErrorMitigationOptions(BaseModel):
    debiasing: Optional[bool] = None


class JobType(str, Enum):
    circuit = "ionq.circuit.v1"
    multi_circuit = "ionq.multi-circuit.v1"
    quantum_function = "quantum-function"
    optimization = "optimization"


class Ansatz(BaseModel):
    format: str = "qasm"
    data: str
    num_qubits: int

    def __init__(self, **data):

        match = re.search(r"qubit\[(\d+)\]", data["data"])
        if match is None:
            raise ValueError("Invalid QASM: No qubit declaration found")

        data["num_qubits"] = int(match.group(1))

        try:
            parser.parse(data["data"])
        except Exception as e:
            raise ValueError(f"Invalid QASM: {e}")

        super().__init__(**data)


class HamiltonianPauliTerm(BaseModel):
    pauli_string: str
    coefficient: float


@dataclass
class Hamiltonian:
    terms: List[HamiltonianPauliTerm]
    num_qubits: Optional[int]


class QuantumFunction(BaseModel):
    pass


class HamiltonianEnergyData(BaseModel):
    hamiltonian: List[HamiltonianPauliTerm]
    ansatz: Ansatz


T = TypeVar("T")


@dataclass
class WorkloadInput(Generic[T]):
    type: JobType
    input: T


class QuantumFunctionWorkload(WorkloadInput[QuantumFunction]):
    pass


class QuantumFunctionInput(BaseModel):
    data: QuantumFunction
    params: Optional[List[float]] = None


class HamiltonianEnergyQuantumFunction(QuantumFunction):
    type: str = Field(default="hamiltonian-energy", const=True)
    data: HamiltonianEnergyData


class CircuitData(BaseModel):
    qasm: Optional[str] = None
    ore: Optional[Any] = None

    def __init__(self, **data):
        super().__init__(**data)
        if self.qasm is None and self.ore is None:
            raise ValueError("Circuit must have qasm or ore")


class CircuitType(str, Enum):
    ore = "ore"
    qis = "qis"
    native = "native"


class OptimizationMethod(str, Enum):
    SPSA = "SPSA"
    Powell = "Powell"
    CG = "CG"
    BFGS = "BFGS"
    Newton_CG = "Newton-CG"
    L_BFGS_B = "L-BFGS-B"
    TNC = "TNC"
    COBYLA = "COBYLA"
    COBYQA = "COBYQA"
    SLSQP = "SLSQP"
    TRUST_CONSTR = "trust-constr"
    DOGLEG = "dogleg"
    TRUST_NCG = "trust-ncg"
    TRUST_EXACT = "trust-exact"
    TRUST_KRYLOV = "trust-krylov"


class OptimizationWorkload(BaseModel):
    quantum_function: QuantumFunction
    method: OptimizationMethod
    initial_params: Optional[List[float]] = None  # temporary setter
    log_interval: Optional[int] = 1
    options: Optional[Dict[str, Any]] = None
    maxiter: Optional[int] = None


class JobConfig(BaseModel):
    target: TargetOptions
    shots: Optional[int] = 100
    noise: Optional[NoiseOptions] = None
    error_mitigation: Optional[ErrorMitigationOptions] = None

    def __init__(self, **data):
        super().__init__(**data)
        if (
            (self.target == TargetOptions.simulator and self.noise is not None)
            or (self.target != TargetOptions.simulator)
            and self.shots is None
        ):
            raise ValueError("Shots must be set for ideal simulator jobs")


class CircuitWorkload(BaseModel):
    gateset: Optional[GateSet] = GateSet.qis
    # TODO Maybe improve typing here
    circuit: List[Dict[str, Any]]
    qubits: Optional[int] = None


class MultiCircuitWorkload(BaseModel):
    gateset: Optional[GateSet] = GateSet.qis
    # TODO Improve typing here
    circuits: List[
        Dict[str, Any]
    ]  # Union[List[CircuitList], List[CircuitOperation], None] = None
    qubits: Optional[int] = None

    # def __init__(self, **data):
    #     super().__init__(**data)
    #     if self.qubits == 0:
    #         raise ValueError("Circuit must have at least one qubit")


class JobSettings(BaseModel):
    error_mitigation: Optional[ErrorMitigationOptions] = None


class JobRequest(BaseModel, Generic[T]):
    name: Optional[str] = None
    metadata: Optional[dict] = None
    shots: Optional[int] = 100
    backend: str
    noise: Optional[NoiseOptions] = None
    type: JobType
    input: T
    settings: Optional[JobSettings] = None

    class Config:
        schema_extra = {
            "name": "My Awesome Job",
            "metadata": {"custom_key": "a string, maximum 400 chars"},
            "shots": 123,
            "backend": "simulator",
            "noise": {"model": "aria-1", "seed": 1},
            "type": "ionq.circuit.v1",
            "input": {
                "qubits": 1,
                "circuits": [{"circuit": {"gate": "h", "target": 0}}],
            },
            "settings": {
                "error_mitigation": {"debiasing": True},
            }
        }


class Progress(BaseModel):
    job_id: str
    status: JobStatus
    finished_at: Optional[str]
    value: Optional[float]
    params: Optional[List[float]]
    iteration: Optional[int]


class JobProgress(BaseModel):
    progress: List[Progress]


class Solution(BaseModel):
    minimum_value: float
    optimal_parameters: List[float]


class QuantumFunctionResult(BaseModel):
    value: Optional[float] = None
    variance: Optional[float] = None


class OptimizationResult(BaseModel):
    solution: Solution


class JobStats(BaseModel):
    gate_counts: Optional[Dict[str, int]] = {}
    qubits: Optional[int] = None
    circuits: Optional[int] = None


class JobDetails(BaseModel):
    id: str
    type: JobType
    name: Optional[str]
    status: JobStatus
    backend: Optional[TargetOptions] = TargetOptions.simulator
    noise: Optional[NoiseOptions]
    metadata: Optional[Dict[str, Union[str, int]]]
    shots: Optional[int] = 100
    settings: Optional[JobSettings]
    stats: Optional[JobStats] = JobStats()
    cost_usd: Optional[float]
    submitted_at: str
    started_at: Optional[str]
    completed_at: Optional[str]
    execution_duration_ms: Optional[int]
    predicted_execution_duration_ms: Optional[int]
    child_job_ids: Optional[List[str]]
    # TODO: type this ?
    results: Optional[Dict[str, Any]]
    failure: Optional[Dict[str, str]]
    warning: Optional[Dict[str, List[str]]]

    class Config:
        schema_extra = {
            "id": "aa54e783-0a9b-4f73-ad2f-63983b6aa4a8",
            "name": "My Awesome Job",
            "status": "completed",
            "target": "qpu.harmony",
            "noise": {"model": "harmony", "seed": 100},
            "metadata": {"custom_key": "a string, maximum 400 chars"},
            "shots": 123,
            "settings": {"error_mitigation": {"debiasing": True}},
            "stats": {
                "gate_counts": {"1q": 8, "2q": 2},
                "qubits": 4,
            },
            "submitted_at": "2025-10-29T19:24:12.280Z",
            "started_at": "2025-10-29T19:24:27.963Z",
            "completed_at": "2025-10-29T19:25:17.153Z",
            "execution_duration_ms": 13,
            "predicted_execution_duration_ms": 13,
            "child_job_ids": ["aa54e783-0a9b-4f73-ad2f-63983b6aa4a8"],
            "results": {
                "shots": {
                    "url": "/v0.4/jobs/aa54e783-0a9b-4f73-ad2f-63983b6aa4a8/results/shots"
                },
                "histogram": {
                    "url": "/v0.4/jobs/aa54e783-0a9b-4f73-ad2f-63983b6aa4a8/results/histogram"
                },
                "probabilities": {
                    "url": "/v0.4/jobs/aa54e783-0a9b-4f73-ad2f-63983b6aa4a8/results/probabilities"
                }
            },
            "failure": {"error": "An error occurred!", "code": "internal_error"},
            "warning": {
                "messages": [
                    "Warning message 1",
                    "Warning message 2",
                    "etc.",
                ]
            },
            "circuits": 1,
        }


class JobList(BaseModel):
    jobs: List[JobDetails]
    next: Optional[str]

    class Config:
        schema_extra = {
            "jobs": [
                {
                    "id": "aa54e783-0a9b-4f73-ad2f-63983b6aa4a8",
                    "name": "My Awesome Job",
                    "status": "completed",
                    "target": "qpu.harmony",
                    "noise": {"model": "harmony", "seed": 100},
                    "metadata": {"custom_key": "a string, maximum 400 chars"},
                    "shots": 123,
                    "error_mitigation": {"debias": True},
                    "gate_counts": {"1q": 8, "2q": 2},
                    "qubits": 4,
                    "cost_usd": 12.41,
                    "request": 1490932820,
                    "start": 1490932821,
                    "response": 1490932834,
                    "execution_time": 13,
                    "predicted_execution_time": 13,
                    "children": ["aa54e783-0a9b-4f73-ad2f-63983b6aa4a8"],
                    "results_url": "/v0.3/jobs/617a1f8b-59d4-435d-aa33-695433d7155e/results",
                    "failure": {
                        "error": "An error occurred!",
                        "code": "internal_error",
                    },
                    "warning": {"messages": ["<string>"]},
                    "circuits": 1,
                }
            ],
            "next": "3c90c3cc-0d44-4b50-8888-8dd25736052a",
        }


class JobUpdate(BaseModel):
    id: str
    ids: Optional[List[str]]
    status: JobStatus

    class Config:
        schema_extra = {
            "id": "aa54e783-0a9b-4f73-ad2f-63983b6aa4a8",
            "status": "ready",
        }

    def __init__(self, **data):
        super().__init__(**data)
        if self.id is None and self.ids is None:
            raise ValueError("Either 'id' or 'ids' must be set")



from __future__ import annotations

import time
import urllib.parse
from typing import Optional, List, Dict, Any
import json
import gzip
import base64

from ..resource import Resource
from ...schemas import JobDetails, JobStatus, JobUpdate, JobRequest, JobProgress
from ...errors import UserCancelledError


def decompress(data: str):
    return json.loads(gzip.decompress(base64.b64decode(data)).decode("utf-8"))


def is_compressed(data: str) -> bool:
    try:
        decompress(data)
        return True
    except Exception:
        return False


class JobsResource(Resource):

    def wait_for_results(self, id: str, polling_rate: float = 1.0) -> Dict[str, Any]:

        while True:
            job = self.get_job(id)  # Get the job and store it in a variable
            if job.status in (
                JobStatus.completed,
                JobStatus.failed,
                JobStatus.canceled,
            ):
                return job.results or {}
            time.sleep(polling_rate)

    def get_jobs(
        self,
        id: Optional[List[str]] = None,
        status: Optional[str] = None,
        limit: int = 25,
        next: Optional[str] = None,
    ) -> List[JobDetails]:
        query_params = {
            key: value
            for key, value in {
                "id": id,
                "status": status,
                "limit": limit,
                "next": next,
            }.items()
            if value is not None
        }
        response = self.client.request(
            "/jobs",
            method="GET",
            params=query_params,
        )

        return [JobDetails(**job) for job in response.get("jobs", [])] or []

    def get_job(
        self,
        job_id: str,
        exclude: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
    ) -> JobDetails:
        params = {
            "exclude": ",".join(exclude) if exclude else None,
            "include": ",".join(include) if include else None,
        }
        response = self.client.request(
            f"/jobs/{job_id}",
            method="GET",
            params=params,
        )

        job = JobDetails(**response)
        if isinstance(job.metadata, dict):
            for key, value in job.metadata.items():
                if is_compressed(str(value)):
                    job.metadata[key] = decompress(str(value))
        return job

    def get_job_progress(self, job_id: str) -> JobProgress:
        response = self.client.request(
            f"/jobs/{job_id}/progress",
            method="GET",
        )

        return JobProgress(**response)

    def create_job(self, job_request: JobRequest) -> JobUpdate:
        job_payload = job_request.dict(exclude_none=True)

        response = self.client.request(
            "/jobs",
            method="POST",
            json=job_payload,
        )
        job_data = JobUpdate(**response)

        return job_data

    def delete_many_jobs(self, job_ids: List[str], force: bool = False) -> JobUpdate:
        if not force:
            if (
                not input(
                    f"Are you sure you want to delete {len(job_ids)} jobs? (y/n): "
                )
                .lower()
                .startswith("y")
            ):
                raise UserCancelledError(
                    "The user canceled the bulk deletion operation."
                )
        response = self.client.request(
            "/jobs",
            method="DELETE",
            json={"ids": job_ids},
        )

        return JobUpdate(**response)

    def delete_job(self, job_id: str, force: bool = False) -> JobUpdate:
        if not force:
            if (
                not input("Are you sure you want to delete this job? (y/n): ")
                .lower()
                .startswith("y")
            ):
                raise UserCancelledError("The user canceled the deletion operation.")
        response = self.client.request(
            f"/jobs/{job_id}",
            method="DELETE",
        )

        return JobUpdate(**response)

    def cancel_job(self, job_id: str) -> JobUpdate:
        response = self.client.request(
            f"/jobs/{job_id}/status/cancel",
            method="PUT",
        )

        return JobUpdate(**response)

    def cancel_many_jobs(self, job_ids: List[str]) -> JobUpdate:
        response = self.client.request(
            "/jobs/status/cancel",
            method="PUT",
            json={"ids": job_ids},
        )

        return JobUpdate(**response)

    def get_job_results(self, uri: str) -> Dict[str, Any]:

        def strip_leading_version_path(url: str) -> str:
            base_parts = urllib.parse.urlparse(self.client.base_url)
            base_path = base_parts.path.rstrip('/')

            path = urllib.parse.urlparse(url).path

            return path[len(base_path):]  # "/specific-resource/123"

        response = self.client.request(
            strip_leading_version_path(uri),
            method="GET",
        )

        return response



from .client import Client
from .jobs.backend import Backend

__all__ = ["Client", "Backend"]

try:
    from .jobs import *  # todo specific classes to export

    # __all__ += []  # Add the names of the classes to export here

except ImportError:
    pass



from typing import List, Literal, Tuple, Dict, Any
from qiskit import QuantumCircuit, qasm3
from qiskit.quantum_info import SparsePauliOp
from qiskit.circuit import controlledgate as q_cgates
from qiskit.circuit.library import standard_gates as q_gates
from ionq.jobs import Circuit
from ionq.schemas.job import Ansatz, Hamiltonian, HamiltonianPauliTerm


def to_ansatz(
    ansatz: QuantumCircuit,
):
    return Ansatz(
        format="qasm",
        data=qasm3.dumps(ansatz),
    )


def to_hamiltonian(
    hamiltonian: SparsePauliOp,
) -> Hamiltonian:

    return Hamiltonian(
        [
            HamiltonianPauliTerm(
                pauli_string=pauli_string, coefficient=coefficient.real
            )
            for pauli_string, coefficient in hamiltonian.to_list()
        ],
        hamiltonian.num_qubits,
    )


# IonQ-specific constants and mappings
ionq_basis_gates = [
    "ccx",
    "ch",
    "cnot",
    "cp",
    "crx",
    "cry",
    "crz",
    "csx",
    "cx",
    "cy",
    "cz",
    "h",
    "i",
    "id",
    "mcp",
    "mcphase",
    "mct",
    "mcx",
    "mcx_gray",
    "measure",
    "p",
    "rx",
    "rxx",
    "ry",
    "ryy",
    "rz",
    "rzz",
    "s",
    "sdg",
    "swap",
    "sx",
    "sxdg",
    "t",
    "tdg",
    "toffoli",
    "x",
    "y",
    "z",
]

ionq_api_aliases = {
    "cp": "cz",
    "csx": "cv",
    "mcphase": "cz",
    "cx": "x",  # TODO: replace all controlled gates with their single-qubit counterparts
    "ccx": "x",
    "mcx": "x",
    "mcx_gray": "x",
    "tdg": "ti",
    "p": "z",
    "rxx": "xx",
    "ryy": "yy",
    "rzz": "zz",
    "sdg": "si",
    "sx": "v",
    "sxdg": "vi",
}

multi_target_uncontrolled_gates = (
    q_gates.SwapGate,
    q_gates.RXXGate,
    q_gates.RYYGate,
    q_gates.RZZGate,
)


def _qiskit_circ_to_ionq_circ(input_circuit: QuantumCircuit):
    """Converts Qiskit circuit to IonQ instructions format."""
    output_circuit = []
    num_meas = 0
    meas_map = [None] * len(input_circuit.clbits)
    for instruction, qargs, cargs in input_circuit.data:
        instruction_name = instruction.name
        if instruction_name == "measure":
            meas_map[input_circuit.clbits.index(cargs[0])] = input_circuit.qubits.index(
                qargs[0]
            )
            num_meas += 1
            continue
        if instruction_name == "id":
            continue
        if instruction_name == "barrier":
            continue
        if instruction_name not in ionq_basis_gates:
            raise ValueError(f"Unsupported instruction: {instruction_name}")

        targets = [input_circuit.qubits.index(qargs[0])]
        if isinstance(instruction, multi_target_uncontrolled_gates):
            targets.extend(input_circuit.qubits.index(q) for q in qargs[1:])

        converted = {
            "gate": ionq_api_aliases.get(instruction_name, instruction_name),
            "targets": targets,
        }
        if len(instruction.params) > 0:
            converted["rotation"] = float(instruction.params[0])
        if isinstance(instruction, q_cgates.ControlledGate):
            converted["controls"] = [input_circuit.qubits.index(qargs[0])]
            converted["targets"] = [input_circuit.qubits.index(qargs[1])]

        output_circuit.append(converted)

    return output_circuit, num_meas, meas_map


def to_circuit(
    circuit: List[QuantumCircuit],
    gateset: Literal["qis", "native"] = "qis",
) -> Circuit:
    """Convert a Qiskit circuit to an IonQ-compatible Circuit workload"""
    ionq_circs: List[Tuple[List[Dict[str, Any]], list, str]] = []
    for circ in circuit:
        ionq_circ, _, meas_map = _qiskit_circ_to_ionq_circ(circ)
        ionq_circs.append((ionq_circ, meas_map, circ.name))

    return Circuit(
        gateset=gateset,
        qubits=max(c.num_qubits for c in circuit),
        circuits=[
            {"name": name, "circuit": circuit, "registers": {"meas_mapped": mapping}}
            for circuit, mapping, name in ionq_circs
        ],
    )



from __future__ import annotations

from typing import List, Optional

from ..resource import Resource
from ...schemas import (
    Characterization,
    Backend as BackendSchema,
)


class BackendsResource(Resource):

    def get_backends(self, status: Optional[str] = None) -> List[BackendSchema]:
        params = {"status": status} if status is not None else {}
        response = self.client.request(
            "/backends",
            method="GET",
            params=params,
        )
        assert isinstance(response, list), "Get Backends response is not a list"
        return [BackendSchema(**backend) for backend in response]

    def get_characterizations(
        self,
        backend: str,
        start: Optional[int] = None,
        end: Optional[int] = None,
        limit: int = 10,
        page: int = 1,
    ) -> List[Characterization]:
        assert backend is not None, "Backend name is required"
        params = {"start": start, "end": end, "limit": limit, "page": page}
        response = self.client.request(
            f"/characterizations/backends/{backend}",
            method="GET",
            params=params,
        )
        assert isinstance(
            response, dict
        ), "Get Characterizations response is not a dict"
        return [
            Characterization(**char) for char in response.get("characterizations", [])
        ]

    def get_current_characterization(self, backend: str) -> Characterization:
        assert backend is not None, "Backend name is required"
        response = self.client.request(
            f"/characterizations/backends/{backend}/current",
            method="GET",
        )
        assert isinstance(
            response, dict
        ), "Get Current Characterization response is not a dict"
        return Characterization(**response)

    def get_characterization(self, uuid: str) -> Characterization:
        response = self.client.request(
            f"/characterizations/{uuid}",
            method="GET",
        )
        assert isinstance(response, dict), "Get Characterization response is not a dict"
        return Characterization(**response)
